;;
;; ep.el
;; emulating My old text formatter, p.exe.
;;
;; Author: Seiichi Namba (C)1994, 1995, 1996 (but copy freely).
;;
;;	Tue Jun 11 21:36:22 1996 version 2.1 ep-forward-sexp macro
;;	Sat Jul 29 14:36:20 1995 version 2.0 Decided to update version to 2.0
;;	Tue Jun 27 04:04:46 1995 version 1.5 Use proprietary macro
;;                                           instead of user commands.
;;                                           True 'Local variable' section
;;                                           support for formatting buffer.
;;	Wed Jun 15 19:58:43 1994 version 1.4 LocalVariable support.
;;					     .ep (startup) file support.
;;	Fri Mar 11 20:19:52 1994 version 1.3 Printer ctrl sequence dictionary.
;;	Thu Mar 10 18:54:06 1994 version 1.2 Explicit FormFeed control.
;;					      -> This made codes really dirty.
;;	Sun Mar  6 20:00:12 1994 version 1.1 Printer control.
;; 	Thu Mar  3 19:58:34 1994 version 1.0 Basic structure.
;;

;; Some macros, 
;; proprietary versions of functions without `ep-' prefix.
;; 
;; Main purpose of these macro is not only speeding things up, but
;; also covering prototype codes in early version with eliminating
;; annoying messages like "Mark set" from GNU Emacs.
;;
;; However, these defmacro versions ep-* were not found to be any better
;; than defun versions for speeding whole process (probably because
;; all of them are not called repeatedly).
;;
(defmacro ep-replace-regexp (re s)
  ;; proprietary version of replace-regexp no message.
  (` (while (re-search-forward (, re) nil t)
    (replace-match (, s) nil nil))))
(defmacro ep-replace-string (s1 s2)
  ;; proprietary version of replace-string, no message.
  (` (while (search-forward (, s1) nil t)
    (replace-match (, s2) nil nil))))
(defmacro ep-insert-buffer-substring (buf rb re)
  ;; proprietary version of insert-buffer-substring, no message.
  (` (insert
      (save-excursion
	(set-buffer (, buf)) (buffer-substring (, rb) (, re))))))
(defmacro ep-insert-buffer (buf)
  ;; proprietary version of insert-buffer, no message.
  (` (insert
      (save-excursion
	(set-buffer (, buf)) (buffer-substring (point-min) (point-max))))))
;;
;; Retain compatibility of forward-sexp
;;
(if (< (string-to-int emacs-version) 19)
    (fset 'ep-forward-sexp 'forward-sexp)
  (defmacro ep-forward-sexp (arg)
    (` (let* ((old-syntax (syntax-table))) ; save old syntax
	 (set-syntax-table ep-local-syntax-table)
	 (forward-sexp (, arg))
	 (set-syntax-table old-syntax)
	 nil))))
(defvar ep-local-syntax-table
  [0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
     0 1 7 1 2 2 3 1 10500 10245 3 3 1 3 1 3 2 2 2 2 2 2 2 2 2 2 1 1 3 3 3 1
     1 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 23812 9 23301 1 3
     1 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 32004 3 31493 1 0
     1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
     1 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
     13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
     13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
     13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
     13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
     13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 1]
  "Copied from (standard-syntax-table) output in emacs 18")

;; p.exe's help message.
;;
;;'p' Kanji print formatter v4.94 pl
;;Usage:
;;p [-t<N> -m<N> -l<N> -c<N> -e<N> -s<N> -S<N> -h <hdr>
;;   -f -P -p -n -d -b -k -W -C -?] [file...]
;;	-t<N>:	Top margin set to N (default 4)
;;	-m<N>:	left Margin set to N (default 16)
;;	-l<N>:	Lines in a page set to N (default 27)
;;	-c<N>:	Characters in a line set to N (default 60)
;;	-e<N>:	Entab width set to N (default 8)
;;	-s<N>:	Start from page N (default 1)
;;	-S<N>:	Stop after printing page N 
;;	-h [h]:	adds user Header 'h'
;;	-f:	line Filling mode
;;	-P:	enable Paragraph # printing (default off)
;;	-p:	disable Page number printing (default on)
;;	-n:	disable line Number printing (default on)
;;	-d:	disable Double spacinng (default on)
;;	-b:	page # in Bottom (default at right shoulder)
;;	-i:	disable file Info printing
;;	-k:	disable Kinsoku processing
;;	-C:	wait for paper Change
;;	-?:	prints this HELP

;
; User options emulating above command line option of p.exe:
;
(defvar ep-top-margin		6	"-t")
(defvar ep-left-margin		14	"-m")
(defvar ep-lines		27	"-l")
(defvar ep-chars		58	"-c")
(defvar ep-tab-width		8	"-e")
(defvar ep-start-page		nil	"-s")
(defvar ep-end-page		nil	"-S")
(defvar ep-user-header		nil	"-h")
(defvar ep-fill-mode		nil	"-f")
(defvar ep-enable-paranum	nil	"-P")
(defvar ep-disable-pnum		nil	"-p")
(defvar ep-disable-lnum		nil	"-l")
(defvar ep-disable-dspc		nil	"-d")
(defvar ep-bottom-pnum		nil	"-b")
(defvar ep-disable-finfo	nil	"-i")
(defvar ep-kinsoku-process	nil	"-k")
(defvar ep-wait-for-pchange	nil	"-C")
(defvar ep-print-help		nil	"-?")	;; NOP now.
(defvar ep-bottom-margin	4)		;; This is new in ep.

;
; User options
;
(defvar ep-dos-style-j-file	nil
  "This option assumes a file like *.jxw file of Ichitarou v3.")
(defvar ep-j-file		t
  "This option controls filling mode with ep-dos-style-j-file")
(defvar ep-enable-justify	nil
  "Meaningless if you are formatting Japanese.")
(defvar ep-enable-all-lnum	nil
  "If non-nil, put line number to all lines.")
(defvar ep-print-current-time nil
  "If non-nil, prints current time instead of default file's modification time.")
(defvar ep-hack-local-variables-always t
  "If non-nil, Local Variable section is always evaluated.")
(defvar ep-reload-startup-file-always nil
  "If non-nil, startup file is always reloadded after formatting.")
(defvar ep-do-not-parse-printer-sequence nil
  "If non-nil, ep does not parse printer controll sequence.
Set to t if 'verbatim' formatting is needed in program source list print.")
(defvar ep-do-not-erase-local-v-section nil
  "If non-nil, ep does not erase Local Variable section.
Set to t if 'verbatim' formatting is needed in program source list print.")
(defvar ep-do-not-display-percentage nil
  "If non-nil, ep does not display percentage.
This speeds formatting up 8-10% approximately.")

;
; Global variables.
;
(defconst ep-buf		"*ep-buf*" "ep's work buffer.")
(defvar ep-header		nil)
(defvar ep-num-width		4 "Width of line # or page # etc. ")
(defvar ep-lnum-to-text		4 "Distance between line # and text.")
(defvar ep-interactive-p	nil "Controls verbose message.")
(defvar ep-startup-file		(if (eq system-type 'ms-dos) "~/_ep" "~/.ep")
  "Startup file. Optional.")

;
; Define formatting parameters as buffer local.
; Is there any other cool way to do this ?
; I don't want to implement
;   (defvar ep-param-alist '((ep-top-margin . 4) ...)
; kind of thing.
;
(make-variable-buffer-local 'ep-top-margin)
(make-variable-buffer-local 'ep-left-margin)
(make-variable-buffer-local 'ep-lines)
(make-variable-buffer-local 'ep-chars)
(make-variable-buffer-local 'ep-tab-width)
(make-variable-buffer-local 'ep-start-page)
(make-variable-buffer-local 'ep-end-page)
(make-variable-buffer-local 'ep-user-header)
(make-variable-buffer-local 'ep-fill-mode)
(make-variable-buffer-local 'ep-enable-paranum)
(make-variable-buffer-local 'ep-disable-pnum)
(make-variable-buffer-local 'ep-disable-lnum)
(make-variable-buffer-local 'ep-disable-dspc)
(make-variable-buffer-local 'ep-bottom-pnum)
(make-variable-buffer-local 'ep-disable-finfo)
(make-variable-buffer-local 'ep-kinsoku-process)
(make-variable-buffer-local 'ep-wait-for-pchange)
(make-variable-buffer-local 'ep-print-help)
(make-variable-buffer-local 'ep-bottom-margin)
(make-variable-buffer-local 'ep-dos-style-j-file)
(make-variable-buffer-local 'ep-j-file)
(make-variable-buffer-local 'ep-enable-justify)
(make-variable-buffer-local 'ep-enable-all-lnum)
(make-variable-buffer-local 'ep-print-current-time)
(make-variable-buffer-local 'ep-hack-local-variables-always)
(make-variable-buffer-local 'ep-reload-startup-file-always)
(make-variable-buffer-local 'ep-do-not-parse-printer-sequence)
(make-variable-buffer-local 'ep-do-not-erase-local-v-section)
(make-variable-buffer-local 'ep-header)
(make-variable-buffer-local 'ep-num-width)
(make-variable-buffer-local 'ep-lnum-to-text)
(make-variable-buffer-local 'ep-do-not-display-percentage)

;
; Printer reset/restore sequence (buffer local for 2.0).
;	12pt
;	mincho
;	Font ctrl (valid only if LaserShot has 40x40 font card).
; These are used by function ep-last-job.
;
;;(defvar ep-printer-reset-sequence "\e?o1\e?T0\e?G40\eT20"
;;  "This is for LaserShot pcpr emulation mode. 20/120 inch linefeed.")
(defvar ep-printer-reset-sequence nil
  "Reset sequence sent to printer at first.")
;;(defvar ep-printer-restore-sequence "\eA"
;;  "This is linefeed width to 1/6 inch for LaserShot pcpr emulation mode.")
(defvar ep-printer-restore-sequence nil
  "Restore sequence sent to printer at last.")

(make-variable-buffer-local 'ep-printer-reset-sequence)
(make-variable-buffer-local 'ep-printer-restore-sequence)

;
; Top level shell (interactive's).
;
(defun ep-buffer ()
  "Formats current buffer and pop up a buffer containing result."
  (interactive)
  (setq ep-interactive-p (interactive-p))
  (ep-region (point-min) (point-max)))

(defun ep-region (rb re)
  "Formats current region and pop up a buffer containing result."
  (interactive "r")
  (or ep-interactive-p	        ;; Because ep-region called both
      (setq ep-interactive-p	;; interactively & noninteractively.
	    (interactive-p)))	;; Dirty ...
  (let ((fname
	 (if (buffer-file-name)
	     (file-name-nondirectory (buffer-file-name))
	   (or (buffer-name)
	       ;; Is ther any buffer without name ? I don't know.
	       "*Anonymous file*"))))

    ;; **** This must be done before ep-header hacking ! ****
    ;;   Eval "Local Variables" section again. I don't want
    ;;   to destroy Emacs convention of eval'ing 'Local Variables'
    ;;   section,  though, we need this feature.
    ;; **** This must be done before ep-header hacking ! ****
    (if ep-hack-local-variables-always
	(hack-local-variables))

    (setq ep-header
	  (if ep-disable-finfo
	      nil
	    (format "%s %s"
		    (if (and (not ep-print-current-time)
			     (buffer-file-name))
			(ep-file-mtime fname)
		      (current-time-string))
		    fname)))

    (run-hooks 'ep-hook)	;; Last minites alignment in source buffer.

    ;; Now switch to formatting buffer, and do the job
    (ep-do-format rb re)

    (run-hooks 'ep-post-hook)	;; Do some cleanups in source buffer.

    ;; reload startup file if user requires.
    (if ep-reload-startup-file-always
	(ep-load-startup-file))

    (display-buffer ep-buf)))	;; Show result.

(defconst ep-plain-buf "*ep-plain-text*"
  "ep's buffer for use in plain text output.")

(defun ep-make-plain-text-buffer (arg)
  "Makes plain text from text in current buffer
with calling ep-preprocess-all-sequences to use its sideeffect (deleteion
of sequences stored in ep-printer-ctrl-seq-dict).
Only formfeed is left undeleted.
If called with any ARG, erases all formfeeds in buffer."
  (interactive "P")
  (let ((cbuf (current-buffer))
	(buf (renew-buffer ep-plain-buf)))
    (save-excursion
      (set-buffer buf)
      (ep-insert-buffer cbuf)
      (ep-preprocess-all-sequences)	;; Use sideeffect (deletion of seq.).
      (setq ep-ctrl-seq-markers-list nil)	;; Reset high cost object.
      (if arg (progn (goto-char (point-min))
		     (ep-erase-local-v-section)  ;; Also erase Local V section.
		     (ep-replace-string "\f" ""))))
    (display-buffer buf)))

(defun ep-stdout ()
  "For batch mode use. Prints *ep-buf* to stdout.
Guess works well on Emacs 19, because V18 Emacs does not distinguish stdin/out."
  ;;(interactive)
  (save-excursion
   (set-buffer ep-buf)
   ;; (message "Is this STTERR?")
   (princ (buffer-substring (point-min) (point-max)) t)))

;
; Subroutines.
;

; ep-do-format, ep-setup-format-buffer, ep-get-local-variable and
; ep-put-local-variable are new in version 1.5. To support copying local
; variables from source buffer to formatting buffer.

(defun ep-do-format (rb re)
  "Top level shell of non-interactive functions in formatting buffer (*ep-buf*).
Initialize ep-buf, feed REGBEG REGEND to ep-format-buffer."
  (let ((cbuf (current-buffer))
	;; ep-blv, ep-comment-start: These may not be seen as bound
	;; globally from a view point of C programmer.  Convenient but
	;; dangerous.  What a anarchy world of lisp. Anyway, ep-blv
	;; KEEPS source buffers' "ep-" related local variables as
	;; alist. ep-comment-start keeps comment-start for deletion of
	;; 'Local Variables' section.
	(ep-blv (ep-get-local-variable))
	(ep-comment-start comment-start))
    (renew-buffer ep-buf)
    ;; Now switch to formatting buffer (ep-buf).
    (save-excursion
      (set-buffer ep-buf)
      (ep-insert-buffer-substring cbuf rb re)
      (ep-setup-format-buffer) ;; Last minites alignment.
      (ep-format-buffer))))

;; Dummies for Demacs 1.2.0. 
;; Demacs sometimes barfs without this.
;; I don't know what I am doing.
(if (eq system-type 'ms-dos)
    (progn
      (defvar ep-comment-start nil)
      (defvar ep-blv nil)))

(defun ep-setup-format-buffer ()
  "Fiddling formatting buffer."
  ;; This resets local variables.
  ;; Must be done here (before ep-put-local-variable hack).
  (fundamental-mode)	      ;; Control sequence handler needs this.

  ;; ep-blv is available now.
  ;; Only ep- related things are copied to ep-buf
  (ep-put-local-variable ep-blv)

  (goto-char (point-min))
  (ep-erase-local-v-section)  ;; Erase "Local Variables" section.
  )

; 'while' version. Which do you like ?  Might be little bit faster.
; C programmer won't be bothered by this kinda stuff.
; Looks like ordinary (what's that ?) computer language.
;(defun ep-get-local-variable ()
;  "Returns alist of buffer local variable starting with 'ep-'."
;  (let ((lis (buffer-local-variables (current-buffer))) out tmp)
;    (while lis
;      (setq tmp (car lis))
;      (if (string-match "^ep-" (symbol-name (car tmp))) 
;	  (setq out (append out (list tmp))))
;      (setq lis (cdr lis)))
;    out))
; mapcar version. Guess rather (6-8%) slower than while (Does it matter ?).
; Easy to read to lisper (provided that reader knows all about mapcar).
; I learned mapcar already (am i lisper now ?).
(defun ep-get-local-variable ()
  "Returns alist of buffer local variable starting with 'ep-'."
  ;; mapcar generates nil entry so use while
  (let (out)
    (mapcar
     (function
      (lambda (x)
	(if (string-match "^ep-" (symbol-name (car x))) 
	    ; I don't want nil element
	    (setq out (append out (list x))))))
     (buffer-local-variables (current-buffer)))
    out))

; This copies stuffs like  buffer-file-name, that's annoying.
;(fset 'ep-get-local-variable 'buffer-local-variables)

(defun ep-put-local-variable (alv)
  "Takes ALV (alist of variable binding) and apply each binding
as buffer local variable in current buffer."
  (mapcar
   (function
    (lambda (x)
      (let ((e1 (car x))
	    (e2 (cdr x)))
;	(make-local-variable e1)
	(set e1 e2)))) alv))

(defun ep-format-buffer ()
  "Formats current buffer. Used by ep-buffer."
  ;; Now the text to be formatted is in current buffer, and we are
  ;; sitting on (point-min). Here we go.

;; Guess Older version needed this. Keep this a little while.
  ;; Only this should be compensated
  ;; because this is already determined in ep-region.
;;  (if ep-disable-finfo 
;;      (setq ep-header nil))

  ;;;;;;;; Physical formatting starts here. ;;;;;;;;

  (setq tab-width ep-tab-width)
  (untabify			;; Point does not move.
   (point-min) (point-max))

  (if (eq system-type 'ms-dos)	;; Erase farty ^Z's.
      (save-excursion (ep-replace-string "\032" "")))

  (if ep-dos-style-j-file
      (progn 
	(setq ep-j-file t)
	;;(ep-lf+)		;; Old code.  New one is:
	(ep-double-space)))	;; Preprocess for filling.

  ;; Remember and mark seq and delete. Should be before fill.
  (or ep-do-not-parse-printer-sequence
      (ep-preprocess-all-sequences))

  (if ep-fill-mode
      (ep-fill-buffer))		;; First fill them all.

  (if (or ep-j-file (not ep-disable-dspc))
      (ep-single-space))	;; preprocess to single space if j-file or double space.

  ;; Insert printer ctrl sequences.
  (or ep-do-not-parse-printer-sequence
      (ep-insert-all-sequences))

  (ep-make-pages)		;; Insert formfeeds at appropriate spots.

  (ep-horizontal-format)	;; Left margin and lnum.

  (or ep-disable-dspc		;; Double spacing.
      (ep-double-space))

  (ep-vertical-format)		;; Make header and top margin.

  (ep-last-job))

; Old (no comment handling), forward-search version.
;(defun ep-erase-local-v-section ()
;  "Erases Local Variables section in the buffer."
;  (save-excursion
;    (let ((case-fold-search t))
;      (if (re-search-forward "^[ \t]*local[ \t]+variables:" nil t nil)
;	  (progn (beginning-of-line nil)		 ;; From BOL.
;		 (delete-region (point) (point-max)))))))
; New one. Supports comment, uses backward-search.
(defun ep-erase-local-v-section ()
  "Erases Local Variables section in the buffer."
  (if ep-do-not-erase-local-v-section
      ()
    (save-excursion
      (let ((case-fold-search t))
	(goto-char (point-max))
	(if (re-search-backward
	     (format "^\%s%s"
		     (if ep-comment-start
			 (concat ep-comment-start "+[ \t]*") "")
		     "local[ \t]+variables:") nil t nil)
	    (progn (beginning-of-line nil)	;; From BOL.
		   (delete-region (point) (point-max))))))))

(defun renew-buffer (bf)
  "Get-buffer-create an absolutely new BUFFER.
Old buffer object and its content are destroyed."
  (if (get-buffer bf) (kill-buffer bf))
  (get-buffer-create bf)
  (save-excursion
      (set-buffer bf)
      (widen)
      ;(erase-buffer)
      (current-buffer)))

; Rough.
;(defun ep-fill-buffer-junk ()
;  (setq fill-column ep-chars)
;  (message "Filling buffer...")
;  (fill-region (point-min) (point-max))
;  (message "done."))

(defun ep-fill-buffer ()
  "Fill, kinsoku, or justification.  Just calls fill-paragraph repeatedly."
  (setq fill-column ep-chars)
  (goto-char (point-min))
  (while (< (point) (point-max))
    (fill-paragraph ep-enable-justify)
    (ep-percent "Filling buffer")
    (forward-paragraph 1)))

(defun ep-double-space ()
  "Really cheap formatting to double space."
  (goto-char (point-min))
  (ep-replace-string "\n"  "\n\n"))

(defun ep-single-space ()
  "Really cheap formatting to single space."
  (goto-char (point-min))
  (ep-replace-string "\n\n" "\n"))

(defun ep-make-pages ()
  "Makes pages in buffer, inserting formfeeds per ep-lines, or at explicit ff."
  (goto-char (point-min))
  (while (< (point) (point-max))
    (ep-goto-eop)
    (insert "\f"))
  (backward-delete-char 1)) ;; Erase last formfeed to avoid making page there.

;
; To support user's explicit FF input. This is the dirtiest part.
;
(defun ep-goto-eop ()
  "Jump ep-lines, or next formfeed if it is prior to the point ep-lines ahead."
  ;; forward-line returns the count of lines left to move.
  ;; What a smart implementation.
  (let (the-rest eop eop-line)
    (setq eop (save-excursion
		(setq the-rest (forward-line ep-lines)
		      eop-line (count-lines (point-min) (point)))
		(point)))
    (if (search-forward "\f" (1+ eop) t nil) ;; 1+, to search FF on border too!
	(progn (backward-delete-char 1)	;; erase the explicit FF and stop.
	       ;; the-rest must be increased by lines to make one page.
	       (setq the-rest
		     (+ the-rest    ;; This is 0 unless early EOF was detected.
			(- eop-line (count-lines (point-min) (point))))))
      ;; In case no explicit FF.
      (goto-char eop))
    ;; If no footer, don't have to make pages by newlines.
    (if ep-bottom-pnum (newline the-rest))))

(defun ep-horizontal-format ()
  "Makes left margin, and line # unless ep-disable-lnum."
  (goto-char (point-min))
  (let ((line 1) (newpage nil))
    (while (< (point) (point-max))
      (if (looking-at "^\f")
	  (progn
	    (forward-char 1)
	    (setq line 1
		  newpage t)))
      (ep-make-left-margin line)
      (forward-line 1)
      (ep-percent "Formatting horizontally")
      (if newpage
	  (setq newpage nil))
      (setq line (1+ line)))))

(defun ep-make-left-margin (line)
  "Makes left margin for a line having line number NUM. Used by ep-horizontal-format."
  (insert
   (if (and (not ep-disable-lnum)
	    (or ep-enable-all-lnum (= (% line 5) 0)))
       (concat (make-string
		(- ep-left-margin (+ ep-num-width ep-lnum-to-text)) ? )
	       (format (concat "%" (int-to-string ep-num-width) "d") line)
	       (make-string ep-lnum-to-text ? ))
     (make-string ep-left-margin ? ))))

(defun ep-vertical-format ()
  "Makes top and bottom margin, and header and footer."
  (goto-char (point-min))
  (let ((pnum 1))
    (goto-char (point-min))
    (while (< (point) (point-max))
      (newline 1)
      (ep-make-header pnum)
      (newline (- ep-top-margin 2))	;; '2' is for above two lines.
      (ep-jump-to-bottom)		;; Cursor sits after FF here. So,
      (backward-char nil)		;; move to on FF, adn then:
      (ep-make-footer pnum)
      (forward-char nil)		;; Not to search same FF again.
      (ep-percent "Formatting vertically")
      (setq pnum (1+ pnum)))))

(defun ep-make-header (pnum)
  "Makes header for a page having page number NUM. Used by ep-vertical-format."
  (let ((header-to-pnum (- ep-chars (length ep-header) ep-num-width)))
    (if (< header-to-pnum 0) (ep-error "No room for header."))
    (insert
     (make-string ep-left-margin ? )
     (or ep-header "")
     (make-string header-to-pnum ? )	;; This is alaways printed.
     (if (and (not ep-bottom-pnum)
	      (not ep-disable-pnum))
	 (format
	  (concat "%" (int-to-string ep-num-width) "d") pnum) "")
     ?\n)))

(defun ep-make-footer (pnum)
  "Used by ep-vertical-format."
  (if (and ep-bottom-pnum (not ep-disable-pnum))
      (progn
	(newline (1- ep-bottom-margin))
	(insert (make-string
		 (+ ep-left-margin (- (/ ep-chars 2) (/ ep-num-width 2))) ? )
		(format
		 (concat "%" (int-to-string ep-num-width) "d\n") pnum)))))

(defun ep-jump-to-bottom ()
  "Move cursor on next formfeed."
  (or (search-forward "\f" nil t nil)
      (goto-char (point-max))))

(defun ep-last-job ()
  "Some minor formatting at last."
  (goto-char (point-min))
  (if ep-printer-reset-sequence
      (insert ep-printer-reset-sequence)) ;; Send first printer reset seq.
  (goto-char (point-max))		  ;; Send last formfeed.
  (insert ?\f)
  (if ep-printer-restore-sequence
      (insert ep-printer-restore-sequence))	;; Send printer restore seq.
  (goto-char (point-min)))

(defun ep-percent (str)
  "Messages position in the current buffer as percentage.
This function eats up lot of cycles. 
Setting  variable ep-do-not-display-percentage to `t' disables
the function of this function."
  (or ep-do-not-display-percentage
      (if ep-interactive-p
	  (message "%s %d%%" str
		   (if (or (< (point) 100) (< (buffer-size) 100))
		       (/ (* 100 (point)) (buffer-size))
		     (/ (point) (/ (buffer-size) 100)))))))

(defun ep-load-startup-file ()
  (if (file-readable-p ep-startup-file)       ;; Load startup file.
      (load ep-startup-file)))

;; Overkill ?
(defun ep-error (str)
  ;; reload startup file if user requires.
  (if ep-reload-startup-file-always
      (ep-load-startup-file))
  (error str))

;(defun ep-file-mtime (fn)
;  (let ((tbuf (renew-buffer " *ep-tmp*")))
;    (save-excursion
;      (set-buffer tbuf)
;      (shell-command 
;       (format
;	"diff -ca %s /dev/null | sed -e 's/^.\*\t//' -e 1q" fn) t)
;      (buffer-substring (point-min) (1- (point-max))))))
;(ep-file-mtime "/usr/local/lib/emacs/lisp/skk8.6/SKK-JISYO.JIS2")

(defun ep-file-mtime (file)
  "Returns modified time string of FILE."
;; Output of file-attributes:
;; 4. Last access time, as a list of two integers. (high 16 bit low 16 bit)
;; 5. Last modification time, likewise.
;; 6. Last status change time, likewise.
      (time-to-string (nth 5 (file-attributes file))))

;; We have time-to-string (at least) in Demacs.
;; On Unixe's V18 emacs, we do not have time-to-string function. Why not !
(or (fboundp 'time-to-string)
    (if (>= (string-to-int emacs-version) 19)
	;; After all that pain of wandering around everywhere,
	;; I found V19 current-time-string can handle time-list conversion.
	(fset 'time-to-string 'current-time-string)
      (defun time-to-string (time-list)
	(let ((high (int-to-string (car time-list)))
	      (low (int-to-string (car (cdr time-list))))
	      (tbuf (renew-buffer " *ep-tmp*")))
	  (save-excursion
	    (set-buffer tbuf)
	    ;(erase-buffer)
	    (call-process "ep-ctime" nil t nil high low)
	    (buffer-substring (point-min) (1- (point-max))))))))

; Unused.
;(defun ep-lf+ ()
;  (message "Making double spacing...")
;  (call-process-region (point-min) (point-max) "lf" t t nil "+")
;  (message "Done."))

;
; Printer control sequence handlers.
; Uses marker object extensively.
;
; Basic idea is erasing user put-in sequences, storing all the position
; where ctrl sequence should be inserted in a list of markers, and
; then after filling buffer, insert all physical ctrl sequence which
; can control the printer.
;
; In version 1.1, all the sequence entered in text by user, or
; physical printer ctrl sequences are stored in named variables of
; Emacs Lisp, but I changed to store them in a 'dictionary' variable
; called ep-printer-ctrl-seq-dict.
;
; Tue Jul  4 01:41:18 1995 v1.5 Added face entry.
; Fri Mar 11 20:19:52 1994 v1.3 Changed to printer ctrl sequence dictionary.
; Sun Mar  6 19:59:24 1994 v1.1 Basic functions to support printer control.
;
; Printer control sequence for NEC PCPR (or LaserShot's emulation mode).
;
(defvar ep-printer-ctrl-seq-dict
  ; ctrl beg.,  enable-ctrl,  disable-ctrl, face in emacs 19.
  '(("\\-{"   "\eX"  "\eY" cyan)	; PCPR underscore.
    ("\\b{"   "\e!"  "\e\"" bold)	; PCPR boldface.
    ("\\^{"   "\es1" "\es0" orange)	; PCPR superscript.
    ("\\_{"   "\es2"  "\es0" Plum1))	; PCPR subscript.
    "A list of printer control sequence.  Looks like:
(LIST1 LIST2 LIST3 ...)
where LISTn is structured as:
(USER-INPUT-SEQUENCEn  ENABLE-CTRL-SEQUENCEn  DISABLE-CTRL-SEQUENCEn FACE)
ENABLE-CTRL-SEQUENCEn and DISABLE-CTRL-SEQUENCEn is Printer control string.
USER-INPUT-SEQUENCEn is a string sequence which is entered in text to be
formatted.  Currently USER-INPUT-SEQUENCEn is searched by non-regexp search.
FACE is your favourite face in which special printing text in the mode
specified by a variable my-ep-printer-ctrl-seq-dict. FACE means nothing in
V18 Emaces.")

;; Now this is also buffer local (from v2.0).
(make-variable-buffer-local 'ep-printer-ctrl-seq-dict)

(defvar ep-ctrl-seq-markers-list nil
  "Holds marker list until all printer control sequence is inserted.
Actually this is a list of list each containing marker pair list.
After ep-preprocess-all-sequences execution, this variable would look like:

(((#<marker at 568 in *ep-buf*> #<marker at 882 in *ep-buf*>)
 (#<marker at 974 in *ep-buf*> #<marker at 1190 in *ep-buf*>))
 ((#<marker at 198 in *ep-buf*> #<marker at 203 in *ep-buf*>)) nil)

This example contains three lists. The first one contains two
'marker-pair' which contains enabling and disabling point of the
control sequence.  The second one contains single 'marker-pair', and
the last one has none.  The number, and order of the lists contained
in this variable correspond to that of constant ep-printer-ctrl-seq-dict.")

(defun ep-preprocess-all-sequences ()
  "Erase user put-in sequences, storing all the position as markers.
See document string of ep-printer-ctrl-seq-dict and ep-ctrl-seq-markers-list."
  (ep-move-all-ff-to-bol)
  (let ((lis ep-printer-ctrl-seq-dict))
    (setq ep-ctrl-seq-markers-list nil)
    (while lis
      (setq ep-ctrl-seq-markers-list
	    (append ep-ctrl-seq-markers-list
		    (list (ep-make-sequence-marker-pair-list
			   (car (car lis))))))
      (setq lis (cdr lis)))))

(defun ep-make-sequence-marker-pair-list (sbeg)
  "Calls ep-make-sequence-marker-pair repeatedly seaching SEQUENCEBEGINNING, and
makes a list of sequence marker pair(s). A sequence marker pair contains two
markers for sequence start and end."
  (goto-char (point-min))
  (let ((lis nil) val)
    (while (setq val (ep-make-sequence-marker-pair sbeg))
      (setq lis (append lis (list val))))
    lis))

(defun ep-make-sequence-marker-pair (sbeg)
  "Searchs a SEQUENCESTARTSYMBOL, erase it and corresponding sequence end symbol,
and makes a list containing the beginning and end 'marker' of the sequence.  This
function absolutely depends on ep's sequence structure, such as \-{}
(ep-underscore-sequence).  Point moves to at the end of SEQUENCESTARTSYMBOL
if search succeeded."
  (if (search-forward sbeg nil t nil)
      (let (start end)
	(save-excursion
	  (backward-char nil)
	  (ep-forward-sexp nil)
	  (backward-delete-char 1)
	  (setq end (point-marker)))
	(goto-char (match-beginning 0))
	(delete-region (point) (match-end 0))
	(setq start (point-marker))
	(list start end))
    nil))

(defun ep-insert-all-sequences ()
  "Reads ep-ctrl-seq-markers-list and ep-printer-ctrl-seq-dict,
and inserts printer control sequence stored in ep-printer-ctrl-seq-dict
at the positions stored in ep-ctrl-seq-markers-list."
  (let ((lis ep-ctrl-seq-markers-list)
	(dict ep-printer-ctrl-seq-dict))
    (while lis
      (ep-insert-sequence (car lis)
			  (nth 1 (car dict))
			  (nth 2 (car dict)))
      (setq lis (cdr lis)
	    dict (cdr dict))))
  (setq ep-ctrl-seq-markers-list nil))	;; Clear high-cost object.

(defun ep-insert-sequence (lis enable-ctrl disable-ctrl)
  "At all marker pair contained in LIST, inserts ENABLECTRL and DISABLECTRL."
  (while lis
    (ep-insert-sequence-pair (car lis) enable-ctrl disable-ctrl)
    (setq lis (cdr lis))))

(defun ep-insert-sequence-pair (pair enable-ctrl disable-ctrl)
  "At marker pair contained in a list PAIR, inserts EN-CTRL and DIS-CTRL."
  (let ((beg (car pair))
	(end (car (cdr pair))) l-start l-end)
    (goto-char beg)
    (setq l-start (count-lines (point-min) (point)))
    (insert enable-ctrl)
    (goto-char end)
    (setq l-end (count-lines (point-min) (point)))
    (insert disable-ctrl)
    (if (not (= l-start l-end))
	(ep-ctrl-each-line l-start l-end enable-ctrl disable-ctrl))))

(defun ep-ctrl-each-line (l-start l-end enable-ctrl disable-ctrl)
  "If LINESTART and LINEEND is not equal, inserts EN-CTRL and DIS-CTRL
at each bol and eol between the LINESTART and LINEEND to print left margin
or line number without special printing attribute enabled by EN-CTRL."
  (let ((line l-start))
    (ep-goto-line line)
    (while (<= line l-end)
      (or (= line l-start)
	  (progn (beginning-of-line 1)
		 (insert enable-ctrl)))
      (or (= line l-end)
	  (progn (end-of-line 1)
		 (insert disable-ctrl)))
      (forward-line 1)
      (setq line (1+ line)))))

(defun ep-goto-line (count)
  "Simplified version of goto-line.  Moves cursor to line NUM."
  (goto-char 1)
  (forward-line (1- count)))

(defun ep-move-all-ff-to-bol ()
  "Move all formfeeds to beginning of line."
  (goto-char (point-min))
  (ep-replace-regexp "^\\(.+\\)\f" "\\1\n\f"))

(ep-load-startup-file)      		      ;; Load startup file.

(run-hooks 'ep-load-hook)
(provide 'ep)

;;
;; Appendix: Horizontal and vertical formatting parameters.
;;

;|                                  WIDTH                             |
;|     LM      |                    CHARS                             |
;|             |                                       EP-NUM-WIDTH\  |
;|             |                              |  HEADER-TO-PNUM   |   |_
;|              Fri Mar  4 15:13:11 1994 file.c                    0333
;                                     :                                TM
;........0234.........................................................._
;        |   |                        :
;          \EP-NUM-WIDTH              :                                LINES
;                                     :
;......................................................................-
;                                                                      BM
;                                   1234                               _

;Local Variables:
;eval: (local-set-key "\e$" 'ispell-complete-word)
;ep-enable-all-lnum: t
;ep-left-margin: 3
;ep-lnum-to-text: 1
;ep-num-width: 2
;ep-lines: 50
;ep-chars: 78
;ep-hack-local-variables-always: t
;ep-do-not-parse-printer-sequence: t
;ep-do-not-erase-local-v-section: t
;ep-disable-dspc: t
;ep-fill-mode: nil
;ep-dos-style-j-file: nil
;ep-j-file: nil
;End:
