;;
;; fax.el (oldname:cutpages.el): 
;;    cut buffer into pages and save each page to different files.
;; 
;; By S.Namba
;;
;; Fri Jul 14 14:02:22 1995, do-psconv created for \(ep-\)*psconv-buffer,
;;                           ps-kanji-code
;; Sun Jul  9 18:12:36 1995, Cleanup unused old commands.
;; Mon May 29 01:10:00 1995, 1st draft
;;
(require 'compile)
(require 'ep)

;; Function name fiddling between V18 and V19.
(or (fboundp 'compile-internal)
    (fset 'compile-internal 'compile1))

(defvar cutpage-EOP-string "\014" "End of page string (^L).")
(defvar cutpage-EOP-string-length (length cutpage-EOP-string))
(defvar cutpage-basename "CutPage" "Basename for cut page files.")

(defvar g3-converter-command "mkg3" "command to convert text to g3 file.")
(defvar psconv-command "psconv")
(defvar ps-kanji-code (cond 
		       ((boundp 'MULE) '*euc-japan*unix)
		       ((boundp 'NEMACS) 3)
		       (t nil))
  "Process io coding system for text-ps converter.")

(defvar fax-spool-command "faxspool")


(defun fax-ps (teln)
  (interactive "sFax to (TEL#): ")
  (let ((ps-fname (concat (buffer-file-name) ".ps")))
    (psconv-buffer)			;; No error expected.
    (sit-for 2) (message "")
    (compile-internal (format "%s %s %s" fax-spool-command teln ps-fname)
		      fax-spool-command)
    ))

(defun fax-ep-ps (teln &optional arg)
  "Invoke ep-psconv-buffer, and hand it to fax-spool-command.
Job is done in background"
  (interactive "sFax to (TEL#): \nP")
  (let ((ep-ps-fname (concat (buffer-file-name) ".ep.ps")))
    (ep-psconv-buffer)			;; No error expected.
    (sit-for 2) (message "")
    (compile-internal (format "%s %s %s" fax-spool-command teln ep-ps-fname)
		      fax-spool-command)
    ))

(defun psconv-buffer ()
  "Prepare ps file via just psconv. Writes ps output with fname.ps"
  (interactive)
  (let* ((ps-fname (concat (buffer-file-name) ".ps"))
	 (buf (current-buffer)))
    (save-window-excursion
      (save-excursion
	(renew-buffer "*psconv*")
	(set-buffer "*psconv*")
	;(erase-buffer)
	(ep-insert-buffer buf)
	(do-psconv ps-fname)
	))
    ))

(defun ep-psconv-buffer ()
  "Prepare ps file via ep-buffer and psconv. Writes ps output with fname.ep.ps"
  (interactive)
  (let* ((ep-ps-fname (concat (buffer-file-name) ".ep.ps")))
    (save-window-excursion
      (save-excursion
	;; Don't insert these two sequences
	(let ((ep-printer-reset-sequence nil)
	      (ep-printer-restore-sequence nil))
	  (ep-buffer))
	(set-buffer ep-buf)
	(ep-make-plain-text-buffer nil)
	(set-buffer ep-plain-buf)
	(do-psconv ep-ps-fname)))))

(defun do-psconv (fname)
  ;; do psconv-command, write to fname.
  ;; Should set coding system.
  (cond ((boundp 'MULE)
	 (let ((default-process-coding-system
		 (cons '*autoconv* ps-kanji-code)))
	   (call-process-region (point-min) (point-max)
				psconv-command t t nil)))
	((boundp 'NEMACS)
	 (let ((default-kanji-process-code ps-kanji-code))
	   (call-process-region (point-min) (point-max)
				psconv-command t t nil)))
	(t (call-process-region (point-min) (point-max)
				psconv-command t t nil)))
  ;; set mtime of file.
  (write-region (point-min) (point-max) fname nil t)
  (message "Wrote to %s" fname))


;
; Obsolete
;
(defun Fax-buffer (arg)
  "Prepare g3 file with ep-buffer and g3-converter-command."
  (interactive "P")
  (let* ((ep-fname (concat (buffer-file-name) ".ep")))
    (save-window-excursion
      (save-excursion
	;; Don't insert these two sequences
	(let ((ep-printer-reset-sequence nil)
	      (ep-printer-restore-sequence nil))
	  (ep-buffer))
	(set-buffer ep-buf)
	(ep-make-plain-text-buffer nil)
	(set-buffer ep-plain-buf)
	(write-region (point-min) (point-max) ep-fname nil t)) )
    (if arg
	(compile-internal 
	 (format "%s %s" g3-converter-command ep-fname)
	 ;(concat g3-converter-command " " ep-fname)
			  g3-converter-command))))

(defun fax-cutpages-buffer (arg)
  "Invokes ep-buffer for current buffer, and breaks into page by cutpages. If ARG, run g3-converter-command for each pages."
  (interactive "P")
  ;; Delete all files CutPage.* (including CutPage.*.g3)
  (call-process shell-file-name nil nil nil
		"-c" (concat "rm " cutpage-basename ".\*"))
  (save-window-excursion
    (save-excursion
      ;; Don't insert these two sequences
      (let ((ep-printer-reset-sequence nil)
	    (ep-printer-restore-sequence nil))
	(ep-buffer))
      (set-buffer ep-buf)
      (ep-make-plain-text-buffer nil)
      (set-buffer ep-plain-buf)
      (cutpages-buffer)))

  (if arg
      (compile-internal
       (format "%s %s%s" g3-converter-command cutpage-basename".\*")
       ;(concat g3-converter-command " " cutpage-basename ".\*")
       g3-converter-command)))

(defun cutpages-buffer ()
  "Cuts pages in current buffer, and saves each page to buffer-file-name.<num>."
  (interactive)
  (let ((fname (or (buffer-file-name)
		   (concat default-directory cutpage-basename)))
	(page 1))
    (save-excursion
      (goto-char (point-min))
      (while (< (point) (point-max))
	(write-string-to-file (copy-page)
			      (concat fname "." (int-to-string page)))      
	(setq page (1+ page))
	(if (search-forward cutpage-EOP-string nil t)
	    (if (looking-at "\n")
		(forward-char 1))
	  (goto-char (point-max)))))))

(defun write-string-to-file (str file &optional append)
  "Write STRING to FILE. If optional 3rd arg APPEND, appends the string.
File i/o is done in ps-kanji-code."
  (let ((buf (renew-buffer " *tmp*")))
	(save-excursion
	  (set-buffer buf)
	  ;(erase-buffer)
	  (insert str)
	  (if (boundp 'NEMACS) (setq kanji-fileio-code ps-kanji-code))
	  (if (boundp 'MULE) (setq file-coding-system ps-kanji-code))
	  (write-region (point-min) (point-max) file append))))

(defun copy-page ()
  "Return current page as string.
'Page' is current point to the point before cutpage-EOP-string"
  (interactive)
  (let ((stat (save-excursion
		(if (search-forward cutpage-EOP-string nil t)
		    (- (point) cutpage-EOP-string-length)
		  nil))))
    (buffer-substring (point) (or stat (point-max)))))

(provide 'fax)
